# flake8: noqa: E501
from secrets import token_bytes

import flask
import flask_login
import pytest
from flask_login import LoginManager

from flask_paseto_extended import PasetoCookieSessionInterface


class InvalidCookieClient:
    def __init__(self, app):
        self.app = app

    def __call__(self, environ, start_response):
        invalid_session = "v4.local.WQ8aDTTIbGL9FmTsdBeYoFTciDXl25i5xCbX6W-vpZ342mFyEOg08ghnSGIKqg2lMUGUDvgYly_o0QtGtSCpAl0IFgwLWvrkn9TI_Qfyv1SUbTOCGdPDNxDFlY8JALm9yO_MddM7dCUgrM4M6ofbNh6HiDqhmNr64wbjcdoYTLhsZxgRBVoOwTbtAYWaO-gp7msnGsr2zrJjZcJKEvPcLacHLh3tTWxJoHt41KRThTHlkicf9KHCWJXNA4jM7gWiDBiHdZjMC2_JsUyPDRhnMD1jNbQkakJY0tEAaiXWq9To_fa4BUgDYl4unIl5WQ08ZMa2560"

        environ["HTTP_COOKIE"] = environ.get("HTTP_COOKIE", f"session={invalid_session}")
        return self.app(environ, start_response)


@pytest.fixture(scope="function")
def app():
    app = flask.Flask(__name__)

    login_manager = LoginManager(app)

    # Our mock database.
    users = {"foo@bar.example": {"password": "mysecret"}}

    # Our simple user class
    class User(flask_login.UserMixin):
        pass

    @login_manager.user_loader
    def user_loader(email):
        if email not in users:
            return

        user = User()
        user.id = email
        return user

    @login_manager.request_loader
    def request_loader(request):
        email = request.form.get("email")
        if email not in users:
            return

        user = User()
        user.id = email
        return user

    @login_manager.unauthorized_handler
    def unauthorized_handler():
        return "Unauthorized"

    @app.route("/login", methods=["POST"])
    def login():
        email = flask.request.form["email"]
        if flask.request.form["password"] == users[email]["password"]:
            user = User()
            user.id = email
            flask_login.login_user(user)
            return flask.redirect(flask.url_for("protected"))
        return "Bad login"

    @app.route("/logout")
    def logout():
        flask_login.logout_user()
        return "Logged out"

    @app.route("/protected")
    @flask_login.login_required
    def protected():
        return "Logged in as: " + flask_login.current_user.id

    return app


class TestPasetoCookieSessionInterface:
    """
    Tests for sample code.
    """

    def test_cookie_session(self, app):
        app.secret_key = "super secret string"
        app.session_interface = PasetoCookieSessionInterface()
        with app.test_client() as c:
            res = c.post(
                "/login",
                data={"email": "foo@bar.example", "password": "mysecret"},
                follow_redirects=True,
            )
            assert res.status_code == 200
            res = c.get("/protected")
            assert res.status_code == 200
            assert res.data == b"Logged in as: foo@bar.example"
            res = c.get("/logout", follow_redirects=True)
            assert res.status_code == 200

    @pytest.mark.parametrize(
        "version, key",
        [
            (1, "super secret string"),
            (2, token_bytes(32)),
            (3, "super secret string"),
            (4, "super secret string"),
        ],
    )
    def test_cookie_session_with_paseto_version(self, app, version, key):
        app.secret_key = key
        app.session_interface = PasetoCookieSessionInterface(paseto_version=version)
        with app.test_client() as c:
            res = c.post(
                "/login",
                data={"email": "foo@bar.example", "password": "mysecret"},
                follow_redirects=True,
            )
            assert res.status_code == 200
            res = c.get("/protected")
            assert res.status_code == 200
            assert res.data == b"Logged in as: foo@bar.example"
            res = c.get("/logout", follow_redirects=True)
            assert res.status_code == 200
            assert res.status_code == 200

    def test_cookie_session_without_secret_key(self, app):
        # app.secret_key = "super secret string"
        app.session_interface = PasetoCookieSessionInterface()
        with app.test_client() as c:
            res = c.post(
                "/login",
                data={"email": "foo@bar.example", "password": "mysecret"},
                follow_redirects=True,
            )
            assert res.status_code == 500

    def test_cookie_session_encode_with_invalid_key(self, app):
        app.secret_key = "not 32bytes secret for v2"
        app.session_interface = PasetoCookieSessionInterface(paseto_version=2)
        with app.test_client() as c:
            res = c.post(
                "/login",
                data={"email": "foo@bar.example", "password": "mysecret"},
                follow_redirects=True,
            )
            assert res.status_code == 500

    def test_cookie_session_decode_with_another_key(self, app):
        app.secret_key = "super secret string"
        app.session_interface = PasetoCookieSessionInterface()
        with app.test_client() as c:
            res = c.post(
                "/login",
                data={"email": "foo@bar.example", "password": "mysecret"},
                follow_redirects=True,
            )
            assert res.status_code == 200

        app.secret_key = "another super secret string"
        with app.test_client() as c:
            res = c.get("/protected")
            assert res.status_code == 200
            assert res.data == b"Unauthorized"

    def test_cookie_session_decode_with_invalid_paseto(self, app):
        app.secret_key = "my super secret"
        app.session_interface = PasetoCookieSessionInterface()
        with app.test_client() as c:
            res = c.post(
                "/login",
                data={"email": "foo@bar.example", "password": "mysecret"},
                follow_redirects=True,
            )
            assert res.status_code == 200

        app.wsgi_app = InvalidCookieClient(app.wsgi_app)
        with app.test_client() as c:
            res = c.get("/protected")
            assert res.status_code == 200
            assert res.data == b"Unauthorized"

    @pytest.mark.parametrize(
        "version, msg",
        [
            (0, "Invalid PASETO version: 0"),
            (5, "Invalid PASETO version: 5"),
            (100, "Invalid PASETO version: 100"),
            ("xxx", "Invalid PASETO version: xxx"),
            ({}, "Invalid PASETO version: {}"),
            ([], "Invalid PASETO version: []"),
        ],
    )
    def test_cookie_session_with_invalid_version(self, app, version, msg):
        app.secret_key = "super secret string"
        with pytest.raises(ValueError) as err:
            app.session_interface = PasetoCookieSessionInterface(paseto_version=version)
            pytest.fail("PasetoCookieSessionInterface() should fail.")
        assert msg in str(err.value)



from datetime import datetime, timedelta

from flask import current_app, session
from flask_login import (
    COOKIE_DURATION,
    COOKIE_HTTPONLY,
    COOKIE_NAME,
    COOKIE_SECURE,
    LoginManager,
    user_loaded_from_cookie,
)

from .utils import decode_cookie, encode_cookie

COOKIE_SAMESITE = None
DEFAULT_REMEMBER_COOKIE_PASETO_VERSION = 4


class PasetoLoginManager(LoginManager):
    def init_app(self, app, add_context_processor=True):
        super().init_app(app, add_context_processor)

        self.paseto_version = app.config.get("REMEMBER_COOKIE_PASETO_VERSION", DEFAULT_REMEMBER_COOKIE_PASETO_VERSION)
        if not isinstance(self.paseto_version, int):
            raise TypeError("REMEMBER_COOKIE_PASETO_VERSION must be int")
        if self.paseto_version not in [1, 2, 3, 4]:
            raise ValueError("REMEMBER_COOKIE_PASETO_VERSION must be 1, 2, 3 or 4")

        self._remember_cookie_key = app.config.get("REMEMBER_COOKIE_PASETO_KEY", None)
        if not isinstance(self._remember_cookie_key, (type(None), str)):
            raise TypeError("REMEMBER_COOKIE_PASETO_KEY must be str")

    def _set_cookie(self, response):
        # cookie settings
        config = current_app.config
        cookie_name = config.get("REMEMBER_COOKIE_NAME", COOKIE_NAME)
        domain = config.get("REMEMBER_COOKIE_DOMAIN")
        path = config.get("REMEMBER_COOKIE_PATH", "/")

        secure = config.get("REMEMBER_COOKIE_SECURE", COOKIE_SECURE)
        httponly = config.get("REMEMBER_COOKIE_HTTPONLY", COOKIE_HTTPONLY)
        samesite = config.get("REMEMBER_COOKIE_SAMESITE", COOKIE_SAMESITE)

        if "_remember_seconds" in session:
            duration = timedelta(seconds=session["_remember_seconds"])
        else:
            duration = config.get("REMEMBER_COOKIE_DURATION", COOKIE_DURATION)

        # prepare data
        data = encode_cookie(str(session["_user_id"]), self._remember_cookie_key)

        if isinstance(duration, int):
            duration = timedelta(seconds=duration)

        try:
            expires = datetime.utcnow() + duration
        except TypeError as err:
            raise TypeError("REMEMBER_COOKIE_DURATION must be datetime.timedelta") from err

        # actually set it
        response.set_cookie(
            cookie_name,
            value=data,
            expires=expires,
            domain=domain,
            path=path,
            secure=secure,
            httponly=httponly,
            samesite=samesite,
        )
        return

    def _load_user_from_remember_cookie(self, cookie):
        user_id = decode_cookie(cookie, self._remember_cookie_key)
        if user_id is None:
            return None
        session["_user_id"] = user_id
        session["_fresh"] = False
        user = None
        if self._user_callback:
            user = self._user_callback(user_id)
        if user is not None:
            app = current_app._get_current_object()
            user_loaded_from_cookie.send(app, user=user)
        return user



# flake8: noqa: E501
import flask
import pytest
from flask import jsonify, make_response

from flask_paseto_extended import (
    PasetoIssuer,
    PasetoVerifier,
    current_paseto,
    paseto_required,
)


@pytest.fixture(scope="function")
def app_with_wrong_public_key():
    users = {"foo@bar.example": {"password": "mysecret"}}

    app = flask.Flask(__name__)

    app.config["PASETO_ISS"] = "https://issuer.example"
    app.config["PASETO_PRIVATE_KEYS"] = [
        {
            "version": 4,
            "key": "-----BEGIN PRIVATE KEY-----\nMC4CAQAwBQYDK2VwBCIEILTL+0PfTOIQcn2VPkpxMwf6Gbt9n4UEFDjZ4RuUKjd0\n-----END PRIVATE KEY-----",
        },
    ]
    issuer = PasetoIssuer(app)

    app.config["PASETO_PUBLIC_KEYS"] = [
        {
            "iss": "https://issuer.exmaple",
            "version": 4,
            # The following public key is not derived from the above private key.
            "key": "-----BEGIN PUBLIC KEY-----\nMCowBQYDK2VwAyEAkv4y3wCgwetRuJUt/EKjNJzaTWMKCNcadaGg6obUFdI=\n-----END PUBLIC KEY-----",
        },
    ]
    verifier = PasetoVerifier(app)

    @verifier.token_loader
    def token_loader(request):
        return request.cookies.get("paseto", None)

    @verifier.verification_error_handler
    def verification_error_handler():
        resp = make_response("Unauthorized")
        resp.delete_cookie("paseto", httponly=True)
        return resp

    @app.route("/login", methods=["GET", "POST"])
    def login():
        email = flask.request.form["email"]
        if flask.request.form["password"] != users[email]["password"]:
            return "Bad login"

        token = issuer.issue(payload={"user": {"email": email}}).decode("utf-8")
        resp = flask.redirect(flask.url_for("protected"))
        resp.set_cookie("paseto", token, httponly=True)
        return resp

    @app.route("/logout")
    def logout():
        resp = make_response("Logged out")
        resp.delete_cookie("paseto", httponly=True)
        return resp

    @app.route("/protected")
    @paseto_required()
    def protected():
        # Token verification always fails.
        return jsonify(current_paseto.payload["user"])

    return app


@pytest.fixture(scope="function")
def app_without_verifier():
    users = {"foo@bar.example": {"password": "mysecret"}}

    app = flask.Flask(__name__)

    app.config["PASETO_ISS"] = "https://issuer.example"
    app.config["PASETO_PRIVATE_KEYS"] = [
        {
            "version": 4,
            "key": "-----BEGIN PRIVATE KEY-----\nMC4CAQAwBQYDK2VwBCIEILTL+0PfTOIQcn2VPkpxMwf6Gbt9n4UEFDjZ4RuUKjd0\n-----END PRIVATE KEY-----",
        },
    ]
    issuer = PasetoIssuer(app)

    # app.config["PASETO_PUBLIC_KEYS"] = [
    #     {
    #         "iss": "https://issuer.exmaple",
    #         "version": 4,
    #         "key": "-----BEGIN PUBLIC KEY-----\nMCowBQYDK2VwAyEAHrnbu7wEfAP9cGBOAHHwmH4Wsot1ciXBHwBBXQ4gsaI=\n-----END PUBLIC KEY-----",
    #     },
    # ]
    # verifier = PasetoVerifier(app)

    @app.route("/login", methods=["GET", "POST"])
    def login():
        email = flask.request.form["email"]
        if flask.request.form["password"] != users[email]["password"]:
            return "Bad login"

        token = issuer.issue(payload={"user": {"email": email}})
        resp = flask.redirect(flask.url_for("protected"))
        resp.set_cookie("paseto", token, httponly=True)
        return resp

    @app.route("/logout")
    def logout():
        resp = make_response("Logged out")
        resp.delete_cookie("paseto", httponly=True)
        return resp

    @app.route("/protected")
    @paseto_required()
    def protected():
        # Token verification always fails.
        return jsonify(current_paseto.payload["user"])

    return app


class TestPasetoVerifier:
    """
    Tests for PasetoVerifier.
    """

    def test_verifier(self):
        app = flask.Flask(__name__)
        app.config["PASETO_SKEW"] = 60
        app.config["PASETO_PUBLIC_KEYS"] = [
            {
                "iss": "https://issuer.exmaple",
                "version": 4,
                "key": "-----BEGIN PUBLIC KEY-----\nMCowBQYDK2VwAyEAHrnbu7wEfAP9cGBOAHHwmH4Wsot1ciXBHwBBXQ4gsaI=\n-----END PUBLIC KEY-----",
            },
        ]
        verifier = PasetoVerifier(app)
        assert hasattr(verifier, "token_loader")
        assert callable(verifier.token_loader)
        assert callable(verifier.token_loader_callback)
        assert hasattr(verifier, "verification_error_handler")
        assert callable(verifier.verification_error_handler)
        assert callable(verifier.verification_error_handler_callback)
        assert hasattr(verifier, "_load_and_verify")
        assert callable(verifier._load_and_verify)

        with pytest.raises(NotImplementedError) as err:
            verifier.token_loader_callback(None)
            pytest.fail("token_loader_callback() must fail.")
        assert "token_loader must be defined." in str(err.value)

        with pytest.raises(NotImplementedError) as err:
            verifier.verification_error_handler_callback()
            pytest.fail("verification_error_handler_callback() must fail.")
        assert "verification_error_handler must be defined." in str(err.value)

    def test_verifier_init_app(self):
        app = flask.Flask(__name__)
        app.config["PASETO_SKEW"] = 60
        app.config["PASETO_PUBLIC_KEYS"] = [
            {
                "iss": "https://issuer.exmaple",
                "version": 4,
                "key": "-----BEGIN PUBLIC KEY-----\nMCowBQYDK2VwAyEAHrnbu7wEfAP9cGBOAHHwmH4Wsot1ciXBHwBBXQ4gsaI=\n-----END PUBLIC KEY-----",
            },
        ]
        verifier = PasetoVerifier()
        verifier.init_app(app)
        assert hasattr(verifier, "token_loader")
        assert callable(verifier.token_loader)
        assert callable(verifier.token_loader_callback)
        assert hasattr(verifier, "verification_error_handler")
        assert callable(verifier.verification_error_handler)
        assert callable(verifier.verification_error_handler_callback)
        assert hasattr(verifier, "_load_and_verify")
        assert callable(verifier._load_and_verify)

        with pytest.raises(NotImplementedError) as err:
            verifier.token_loader_callback(None)
            pytest.fail("token_loader_callback() must fail.")
        assert "token_loader must be defined." in str(err.value)

        with pytest.raises(NotImplementedError) as err:
            verifier.verification_error_handler_callback()
            pytest.fail("verification_error_handler_callback() must fail.")
        assert "verification_error_handler must be defined." in str(err.value)

    def test_verifier_init_app_with_paserk(self):
        app = flask.Flask(__name__)
        app.config["PASETO_SKEW"] = 60
        app.config["PASETO_PUBLIC_KEYS"] = [
            {
                "iss": "https://issuer.exmaple",
                "paserk": "k4.public.Hrnbu7wEfAP9cGBOAHHwmH4Wsot1ciXBHwBBXQ4gsaI",
            },
        ]
        verifier = PasetoVerifier()
        verifier.init_app(app)
        assert hasattr(verifier, "token_loader")

    def test_verifier_init_app_with_multiple_paserks(self):
        app = flask.Flask(__name__)
        app.config["PASETO_SKEW"] = 60
        app.config["PASETO_PUBLIC_KEYS"] = [
            {
                "paserk": "k4.public.Hrnbu7wEfAP9cGBOAHHwmH4Wsot1ciXBHwBBXQ4gsaI",
            },
            {
                "paserk": "k3.public.AnBxcnN0dXZ3eHl6e3x9fn-AgYKDhIWGh4iJiouMjY6PkJGSk5SVlpeYmZqbnJ2enw",
            },
        ]
        verifier = PasetoVerifier()
        verifier.init_app(app)
        assert hasattr(verifier, "token_loader")

    def test_verifier_paseto_verification_failed(self, app_with_wrong_public_key):
        with app_with_wrong_public_key.test_client() as c:
            res = c.post(
                "/login",
                data={"email": "foo@bar.example", "password": "mysecret"},
                follow_redirects=True,
            )
            assert res.status_code == 200
            res = c.get("/protected")
            assert res.status_code == 200
            assert res.data == b"Unauthorized"

    def test_verifier_paseto_required_before_setting(self, app_without_verifier):
        with app_without_verifier.test_client() as c:
            res = c.post(
                "/login",
                data={"email": "foo@bar.example", "password": "mysecret"},
                follow_redirects=True,
            )
            assert res.status_code == 500

    @pytest.mark.parametrize(
        "skew, msg",
        [
            (-1, "PASETO_SKEW must be int (>= 0)."),
            (-3600, "PASETO_SKEW must be int (>= 0)."),
            ("3600", "PASETO_SKEW must be int (>= 0)."),
            ([3600], "PASETO_SKEW must be int (>= 0)."),
            ({"value": 3600}, "PASETO_SKEW must be int (>= 0)."),
        ],
    )
    def test_verifier_with_invalid_skew(self, skew, msg):
        app = flask.Flask(__name__)
        app.config["PASETO_SKEW"] = skew
        app.config["PASETO_PUBLIC_KEYS"] = [
            {
                "iss": "https://issuer.exmaple",
                "version": 4,
                "key": "-----BEGIN PUBLIC KEY-----\nMCowBQYDK2VwAyEAHrnbu7wEfAP9cGBOAHHwmH4Wsot1ciXBHwBBXQ4gsaI=\n-----END PUBLIC KEY-----",
            },
        ]
        with pytest.raises(ValueError) as err:
            PasetoVerifier(app)
            pytest.fail("init_app() must fail.")
        assert msg in str(err.value)

    @pytest.mark.parametrize(
        "deserializer, msg",
        [
            (1, "PASETO_DESERIALIZER must have a callable 'loads'."),
            ("string", "PASETO_DESERIALIZER must have a callable 'loads'."),
            ({"loads": ""}, "PASETO_DESERIALIZER must have a callable 'loads'."),
        ],
    )
    def test_verifier_with_invalid_deserializer(self, deserializer, msg):
        app = flask.Flask(__name__)
        app.config["PASETO_SKEW"] = 60
        app.config["PASETO_DESERIALIZER"] = deserializer
        app.config["PASETO_PUBLIC_KEYS"] = [
            {
                "iss": "https://issuer.exmaple",
                "version": 4,
                "key": "-----BEGIN PUBLIC KEY-----\nMCowBQYDK2VwAyEAHrnbu7wEfAP9cGBOAHHwmH4Wsot1ciXBHwBBXQ4gsaI=\n-----END PUBLIC KEY-----",
            },
        ]
        with pytest.raises(ValueError) as err:
            PasetoVerifier(app)
            pytest.fail("init_app() must fail.")
        assert msg in str(err.value)

    @pytest.mark.parametrize(
        "keys, msg",
        [
            ([], "PASETO_PUBLIC_KEYS must be set."),
            (
                [{}],
                "A key object must have a 'paserk' or a pair of 'version' and 'key'.",
            ),
            (
                [{"paserk": "k4.public.xxx"}],
                "Invalid PASERK data.",
            ),
            (
                [{"paserk": "k4.local.b3VyLXNlY3JldA"}],
                "A local key is not allowed.",
            ),
            (
                [{"version": "xxx"}],
                "A 'version' in PASETO_PUBLIC_KEYS must be int.",
            ),
            (
                [{"version": 0}],
                "Invalid PASETO version: 0.",
            ),
            (
                [{"version": 4}],
                "A key object must have a 'paserk' or a pair of 'version' and 'key'.",
            ),
            (
                [{"version": 4, "key": "xxx"}],
                "A 'key' must be a PEM formatted key.",
            ),
            (
                [
                    {
                        "iss": 1,
                        "paserk": "k4.public.Hrnbu7wEfAP9cGBOAHHwmH4Wsot1ciXBHwBBXQ4gsaI",
                    }
                ],
                "An 'iss' must be str.",
            ),
        ],
    )
    def test_verifier_with_invalid_keys(self, keys, msg):
        app = flask.Flask(__name__)
        app.config["PASETO_SKEW"] = 60
        app.config["PASETO_PUBLIC_KEYS"] = keys
        with pytest.raises(ValueError) as err:
            PasetoVerifier(app)
            pytest.fail("init_app() must fail.")
        assert msg in str(err.value)



import json
import typing as t

import flask
from flask import request
from flask.globals import request_ctx
from pyseto import Key, Paseto

from .config import PASETO_VERSIONS_ACCEPTABLE
from .token import Token

PASETO_DEFAULT_SKEW: int = 0
PASETO_DEFAULT_DESERIALIZER: t.Any = json


def _default_token_loader(cb: flask.Request) -> str:
    raise NotImplementedError("token_loader must be defined.")


def _default_verification_error_handler() -> flask.Response:
    raise NotImplementedError("verification_error_handler must be defined.")


class PasetoVerifier:
    def __init__(self, app=None, add_context_processor=True):
        if app is not None:
            self.init_app(app, add_context_processor)

        # Callbacks
        self._token_loader: t.Callable[[flask.Request], str] = _default_token_loader
        self._verification_error_handler: t.Callable[[None], flask.Response] = _default_verification_error_handler

    def init_app(self, app, add_context_processor=True):
        """
        Configures an Flask application to use this PasetoVerifier.
        """
        _ = add_context_processor  # Preserved for drop-in compatibility with LoginManager.
        app.paseto_verifier = self

        # _skew
        self._skew: int = app.config.get("PASETO_SKEW", PASETO_DEFAULT_SKEW)
        if not isinstance(self._skew, int) or self._skew < 0:
            raise ValueError("PASETO_SKEW must be int (>= 0).")

        # _deserializer
        self._deserializer: t.Any = app.config.get("PASETO_DESERIALIZER", PASETO_DEFAULT_DESERIALIZER)
        if not hasattr(self._deserializer, "loads") or not callable(self._deserializer.loads):
            raise ValueError("PASETO_DESERIALIZER must have a callable 'loads'.")

        # _keys
        keys: list = app.config.get("PASETO_PUBLIC_KEYS", [])
        if not keys:
            raise ValueError("PASETO_PUBLIC_KEYS must be set.")
        self._keys: list = []
        self._issuers: dict = {}
        key: t.Any = None
        kid: bytes = b""
        for k in keys:
            if "paserk" in k:
                try:
                    key = Key.from_paserk(k["paserk"])
                except Exception as err:
                    raise ValueError("Invalid PASERK data.") from err
                if key.purpose == "local":
                    raise ValueError("A local key is not allowed.")
            else:
                if "version" not in k:
                    raise ValueError("A key object must have a 'paserk' or a pair of 'version' and 'key'.")
                if not isinstance(k["version"], int):
                    raise ValueError("A 'version' in PASETO_PUBLIC_KEYS must be int.")
                if k["version"] not in PASETO_VERSIONS_ACCEPTABLE:
                    raise ValueError(f"Invalid PASETO version: {k['version']}.")
                if "key" not in k:
                    raise ValueError("A key object must have a 'paserk' or a pair of 'version' and 'key'.")
                try:
                    key = Key.new(k["version"], "public", k["key"])
                except Exception as err:
                    raise ValueError("A 'key' must be a PEM formatted key.") from err
            self._keys.append(key)
            if "iss" in k:
                if not isinstance(k["iss"], str):
                    raise ValueError("An 'iss' must be str.")
                kid = key.to_paserk_id()
                self._issuers[kid] = k["iss"]

        self._paseto = Paseto.new(leeway=self._skew)
        return

    def token_loader(self, cb: t.Callable[[flask.Request], str]) -> t.Callable[[flask.Request], str]:
        """ """
        self._token_loader = cb
        return self.token_loader_callback

    @property
    def token_loader_callback(self):
        """ """
        return self._token_loader

    def verification_error_handler(self, cb: t.Callable[[None], flask.Response]) -> t.Callable[[None], flask.Response]:
        """ """
        self._verification_error_handler = cb
        return self.verification_error_handler_callback

    @property
    def verification_error_handler_callback(self):
        """ """
        return self._verification_error_handler

    def _load_and_verify(self):
        ctx = request_ctx
        token = self._token_loader(request)
        if not token:
            ctx.paseto = Token(
                False,
                error=Exception("A PASETO token could not be loaded via 'token_loader'."),
            )
            return
        try:
            t = self._paseto.decode(self._keys, token, deserializer=self._deserializer)
            ctx.paseto = Token(True, t.version, t.purpose, t.payload, t.footer)
        except Exception as err:
            ctx.paseto = Token(False, error=err)
        return



import pyseto
from flask import current_app
from pyseto import Key


def encode_cookie(payload, key=None):
    if not current_app.secret_key and not key:
        raise ValueError("Either app.secret_key or key should be specified")

    k = key if key else current_app.secret_key
    enc_key = Key.new(current_app.login_manager.paseto_version, "local", k)
    return pyseto.encode(enc_key, payload).decode("utf-8")


def decode_cookie(cookie, key=None):
    k = key if key else current_app.secret_key
    dec_key = Key.new(current_app.login_manager.paseto_version, "local", k)
    return pyseto.decode(dec_key, cookie).payload.decode("utf-8")



# flake8: noqa: E501
import json

import flask
import pyseto
import pytest
from pyseto import Key

from flask_paseto_extended import EncodeError, PasetoIssuer


class TestPasetoIssuer:
    """
    Tests for PasetoIssuer.
    """

    def test_issuer(self):
        app = flask.Flask(__name__)
        app.config["PASETO_ISS"] = "https://issuer.example"
        app.config["PASETO_USE_ISS"] = False
        app.config["PASETO_USE_IAT"] = True
        app.config["PASETO_EXP"] = 3600
        app.config["PASETO_USE_KID"] = True
        app.config["PASETO_SERIALIZER"] = json
        app.config["PASETO_PRIVATE_KEYS"] = [
            {
                "version": 4,
                "key": "-----BEGIN PRIVATE KEY-----\nMC4CAQAwBQYDK2VwBCIEILTL+0PfTOIQcn2VPkpxMwf6Gbt9n4UEFDjZ4RuUKjd0\n-----END PRIVATE KEY-----",
            },
        ]
        issuer = PasetoIssuer(app)
        assert hasattr(issuer, "issue")
        assert callable(issuer.issue)
        token = issuer.issue({"key": "value"})
        assert isinstance(token, bytes)
        key = Key.new(
            4,
            "public",
            "-----BEGIN PUBLIC KEY-----\nMCowBQYDK2VwAyEAHrnbu7wEfAP9cGBOAHHwmH4Wsot1ciXBHwBBXQ4gsaI=\n-----END PUBLIC KEY-----",
        )
        decoded = pyseto.decode(key, token, deserializer=json)
        assert "kid" in decoded.footer

    def test_issuer_with_mandatory_configs(self):
        app = flask.Flask(__name__)
        app.config["PASETO_ISS"] = "https://issuer.example"
        app.config["PASETO_PRIVATE_KEYS"] = [
            {
                "version": 4,
                "key": "-----BEGIN PRIVATE KEY-----\nMC4CAQAwBQYDK2VwBCIEILTL+0PfTOIQcn2VPkpxMwf6Gbt9n4UEFDjZ4RuUKjd0\n-----END PRIVATE KEY-----",
            },
        ]
        issuer = PasetoIssuer(app)
        assert hasattr(issuer, "issue")
        assert callable(issuer.issue)

    def test_issuer_init_app(self):
        app = flask.Flask(__name__)
        app.config["PASETO_ISS"] = "https://issuer.example"
        app.config["PASETO_PRIVATE_KEYS"] = [
            {
                "version": 4,
                "key": "-----BEGIN PRIVATE KEY-----\nMC4CAQAwBQYDK2VwBCIEILTL+0PfTOIQcn2VPkpxMwf6Gbt9n4UEFDjZ4RuUKjd0\n-----END PRIVATE KEY-----",
            },
        ]
        issuer = PasetoIssuer()
        issuer.init_app(app)
        assert hasattr(issuer, "issue")
        assert callable(issuer.issue)

    def test_issuer_with_multiple_keys(self):
        app = flask.Flask(__name__)
        app.config["PASETO_ISS"] = "https://issuer.example"
        app.config["PASETO_PRIVATE_KEYS"] = [
            {
                "version": 4,
                "key": "-----BEGIN PRIVATE KEY-----\nMC4CAQAwBQYDK2VwBCIEILTL+0PfTOIQcn2VPkpxMwf6Gbt9n4UEFDjZ4RuUKjd0\n-----END PRIVATE KEY-----",
            },
            {
                "version": 4,
                "key": "-----BEGIN PRIVATE KEY-----\nMC4CAQAwBQYDK2VwBCIEIGmfHRcqkCfnAOB7234NNeuBpHUVHSLX4z3s4hsaTEQ8\n-----END PRIVATE KEY-----",
            },
        ]
        issuer = PasetoIssuer(app)
        assert hasattr(issuer, "issue")
        assert callable(issuer.issue)

    def test_issuer_init_app_with_paserk(self):
        app = flask.Flask(__name__)
        app.config["PASETO_ISS"] = "https://issuer.example"
        app.config["PASETO_PRIVATE_KEYS"] = [
            {
                "iss": "https://issuer.exmaple",
                "paserk": "k4.secret.cHFyc3R1dnd4eXp7fH1-f4CBgoOEhYaHiImKi4yNjo8c5WpIyC_5kWKhS8VEYSZ05dYfuTF-ZdQFV4D9vLTcNQ",
            },
        ]
        issuer = PasetoIssuer()
        issuer.init_app(app)
        assert hasattr(issuer, "issue")

    def test_issuer_init_app_with_multiple_paserks(self):
        app = flask.Flask(__name__)
        app.config["PASETO_ISS"] = "https://issuer.example"
        app.config["PASETO_PRIVATE_KEYS"] = [
            {
                "paserk": "k4.secret.cHFyc3R1dnd4eXp7fH1-f4CBgoOEhYaHiImKi4yNjo8c5WpIyC_5kWKhS8VEYSZ05dYfuTF-ZdQFV4D9vLTcNQ",
            },
            {
                "paserk": "k3.secret.cHFyc3R1dnd4eXp7fH1-f4CBgoOEhYaHiImKi4yNjo-QkZKTlJWWl5iZmpucnZ6f",
            },
        ]
        issuer = PasetoIssuer()
        issuer.init_app(app)
        assert hasattr(issuer, "issue")

    @pytest.mark.parametrize(
        "iss, msg",
        [
            (None, "PASETO_ISS must be set."),
            ("", "PASETO_ISS must be set."),
            (0, "PASETO_ISS must be set."),
            (1, "PASETO_ISS must be str."),
            (["https://issuer.example"], "PASETO_ISS must be str."),
            ({"iss": "https://issuer.example"}, "PASETO_ISS must be str."),
        ],
    )
    def test_issuer_with_invalid_iss(self, iss, msg):
        app = flask.Flask(__name__)
        app.config["PASETO_ISS"] = iss
        app.config["PASETO_PRIVATE_KEYS"] = [
            {
                "iss": "https://issuer.exmaple",
                "version": 4,
                "key": "-----BEGIN PRIVATE KEY-----\nMC4CAQAwBQYDK2VwBCIEILTL+0PfTOIQcn2VPkpxMwf6Gbt9n4UEFDjZ4RuUKjd0\n-----END PRIVATE KEY-----",
            },
        ]
        with pytest.raises(ValueError) as err:
            PasetoIssuer(app)
            pytest.fail("init_app() must fail.")
        assert msg in str(err.value)

    @pytest.mark.parametrize(
        "use_iss, msg",
        [
            (None, "PASETO_USE_ISS must be bool."),
            ("", "PASETO_USE_ISS must be bool."),
            (b"True", "PASETO_USE_ISS must be bool."),
            ("True", "PASETO_USE_ISS must be bool."),
            (b"False", "PASETO_USE_ISS must be bool."),
            ("False", "PASETO_USE_ISS must be bool."),
            ([True], "PASETO_USE_ISS must be bool."),
            ({"value": True}, "PASETO_USE_ISS must be bool."),
            (100, "PASETO_USE_ISS must be bool."),
        ],
    )
    def test_issuer_with_invalid_use_iss(self, use_iss, msg):
        app = flask.Flask(__name__)
        app.config["PASETO_ISS"] = "https://issuer.example"
        app.config["PASETO_USE_ISS"] = use_iss
        app.config["PASETO_PRIVATE_KEYS"] = [
            {
                "iss": "https://issuer.exmaple",
                "version": 4,
                "key": "-----BEGIN PRIVATE KEY-----\nMC4CAQAwBQYDK2VwBCIEILTL+0PfTOIQcn2VPkpxMwf6Gbt9n4UEFDjZ4RuUKjd0\n-----END PRIVATE KEY-----",
            },
        ]
        with pytest.raises(ValueError) as err:
            PasetoIssuer(app)
            pytest.fail("init_app() must fail.")
        assert msg in str(err.value)

    @pytest.mark.parametrize(
        "use_iat, msg",
        [
            (None, "PASETO_USE_IAT must be bool."),
            ("", "PASETO_USE_IAT must be bool."),
            (b"True", "PASETO_USE_IAT must be bool."),
            ("True", "PASETO_USE_IAT must be bool."),
            (b"False", "PASETO_USE_IAT must be bool."),
            ("False", "PASETO_USE_IAT must be bool."),
            ([True], "PASETO_USE_IAT must be bool."),
            ({"value": True}, "PASETO_USE_IAT must be bool."),
            (100, "PASETO_USE_IAT must be bool."),
        ],
    )
    def test_issuer_with_invalid_use_iat(self, use_iat, msg):
        app = flask.Flask(__name__)
        app.config["PASETO_ISS"] = "https://issuer.example"
        app.config["PASETO_USE_IAT"] = use_iat
        app.config["PASETO_PRIVATE_KEYS"] = [
            {
                "iss": "https://issuer.exmaple",
                "version": 4,
                "key": "-----BEGIN PRIVATE KEY-----\nMC4CAQAwBQYDK2VwBCIEILTL+0PfTOIQcn2VPkpxMwf6Gbt9n4UEFDjZ4RuUKjd0\n-----END PRIVATE KEY-----",
            },
        ]
        with pytest.raises(ValueError) as err:
            PasetoIssuer(app)
            pytest.fail("init_app() must fail.")
        assert msg in str(err.value)

    @pytest.mark.parametrize(
        "exp, msg",
        [
            (-1, "PASETO_EXP must be int (>= 0)."),
            (-3600, "PASETO_EXP must be int (>= 0)."),
            ("3600", "PASETO_EXP must be int (>= 0)."),
            ([3600], "PASETO_EXP must be int (>= 0)."),
            ({"value": 3600}, "PASETO_EXP must be int (>= 0)."),
        ],
    )
    def test_issuer_with_invalid_exp(self, exp, msg):
        app = flask.Flask(__name__)
        app.config["PASETO_ISS"] = "https://issuer.example"
        app.config["PASETO_EXP"] = exp
        app.config["PASETO_PRIVATE_KEYS"] = [
            {
                "iss": "https://issuer.exmaple",
                "version": 4,
                "key": "-----BEGIN PRIVATE KEY-----\nMC4CAQAwBQYDK2VwBCIEILTL+0PfTOIQcn2VPkpxMwf6Gbt9n4UEFDjZ4RuUKjd0\n-----END PRIVATE KEY-----",
            },
        ]
        with pytest.raises(ValueError) as err:
            PasetoIssuer(app)
            pytest.fail("init_app() must fail.")
        assert msg in str(err.value)

    @pytest.mark.parametrize(
        "use_kid, msg",
        [
            (None, "PASETO_USE_KID must be bool."),
            ("", "PASETO_USE_KID must be bool."),
            (b"True", "PASETO_USE_KID must be bool."),
            ("True", "PASETO_USE_KID must be bool."),
            (b"False", "PASETO_USE_KID must be bool."),
            ("False", "PASETO_USE_KID must be bool."),
            ([True], "PASETO_USE_KID must be bool."),
            ({"value": True}, "PASETO_USE_KID must be bool."),
            (100, "PASETO_USE_KID must be bool."),
        ],
    )
    def test_issuer_with_invalid_use_kid(self, use_kid, msg):
        app = flask.Flask(__name__)
        app.config["PASETO_ISS"] = "https://issuer.example"
        app.config["PASETO_USE_KID"] = use_kid
        app.config["PASETO_PRIVATE_KEYS"] = [
            {
                "iss": "https://issuer.exmaple",
                "version": 4,
                "key": "-----BEGIN PRIVATE KEY-----\nMC4CAQAwBQYDK2VwBCIEILTL+0PfTOIQcn2VPkpxMwf6Gbt9n4UEFDjZ4RuUKjd0\n-----END PRIVATE KEY-----",
            },
        ]
        with pytest.raises(ValueError) as err:
            PasetoIssuer(app)
            pytest.fail("init_app() must fail.")
        assert msg in str(err.value)

    @pytest.mark.parametrize(
        "serializer, msg",
        [
            (1, "PASETO_SERIALIZER must have a callable 'dumps'."),
            ("string", "PASETO_SERIALIZER must have a callable 'dumps'."),
            ({"dumps": ""}, "PASETO_SERIALIZER must have a callable 'dumps'."),
        ],
    )
    def test_issuer_with_invalid_serializer(self, serializer, msg):
        app = flask.Flask(__name__)
        app.config["PASETO_ISS"] = "https://issuer.example"
        app.config["PASETO_SERIALIZER"] = serializer
        app.config["PASETO_PRIVATE_KEYS"] = [
            {
                "iss": "https://issuer.exmaple",
                "version": 4,
                "key": "-----BEGIN PRIVATE KEY-----\nMC4CAQAwBQYDK2VwBCIEILTL+0PfTOIQcn2VPkpxMwf6Gbt9n4UEFDjZ4RuUKjd0\n-----END PRIVATE KEY-----",
            },
        ]
        with pytest.raises(ValueError) as err:
            PasetoIssuer(app)
            pytest.fail("init_app() must fail.")
        assert msg in str(err.value)

    @pytest.mark.parametrize(
        "keys, msg",
        [
            ([], "PASETO_PRIVATE_KEYS must be set."),
            (
                [{}],
                "A key object must have a 'paserk' or a pair of 'version' and 'key'.",
            ),
            (
                [{"paserk": "k4.secret.xxx"}],
                "Invalid PASERK data.",
            ),
            (
                [{"paserk": "k4.local.b3VyLXNlY3JldA"}],
                "A local key is not allowed.",
            ),
            (
                [{"version": "xxx"}],
                "A 'version' in PASETO_PRIVATE_KEYS must be int.",
            ),
            (
                [{"version": 0}],
                "Invalid PASETO version: 0.",
            ),
            (
                [{"version": 4}],
                "A key object must have a 'paserk' or a pair of 'version' and 'key'.",
            ),
            (
                [{"version": 4, "key": "xxx"}],
                "A 'key' must be a PEM formatted key.",
            ),
        ],
    )
    def test_issuer_with_invalid_keys(self, keys, msg):
        app = flask.Flask(__name__)
        app.config["PASETO_ISS"] = "https://issuer.example"
        app.config["PASETO_PRIVATE_KEYS"] = keys
        with pytest.raises(ValueError) as err:
            PasetoIssuer(app)
            pytest.fail("init_app() must fail.")
        assert msg in str(err.value)

    def test_issuer_issue_with_bad_kid(self):
        app = flask.Flask(__name__)
        app.config["PASETO_ISS"] = "https://issuer.example"
        app.config["PASETO_PRIVATE_KEYS"] = [
            {
                "version": 4,
                "key": "-----BEGIN PRIVATE KEY-----\nMC4CAQAwBQYDK2VwBCIEILTL+0PfTOIQcn2VPkpxMwf6Gbt9n4UEFDjZ4RuUKjd0\n-----END PRIVATE KEY-----",
            },
            {
                "version": 4,
                "key": "-----BEGIN PRIVATE KEY-----\nMC4CAQAwBQYDK2VwBCIEIGmfHRcqkCfnAOB7234NNeuBpHUVHSLX4z3s4hsaTEQ8\n-----END PRIVATE KEY-----",
            },
        ]
        issuer = PasetoIssuer(app)
        with pytest.raises(ValueError) as err:
            issuer.issue(
                {"foo": "bar"},
                kid="k3.pid.gnwg7IkzZyQF9wJgLLT0OpbdMT7BYmdQoG2u-xXpeeHz",
            )
            pytest.fail("issue() must fail.")
        assert "A signing key is not found." in str(err.value)

    def test_issuer_issue_with_bad_serializer(self):
        class _BadSerializer:
            def dumps(*args, **kwargs):
                raise NotImplementedError("Not implemented.")

        app = flask.Flask(__name__)
        app.config["PASETO_ISS"] = "https://issuer.example"
        app.config["PASETO_SERIALIZER"] = _BadSerializer()
        app.config["PASETO_PRIVATE_KEYS"] = [
            {
                "version": 4,
                "key": "-----BEGIN PRIVATE KEY-----\nMC4CAQAwBQYDK2VwBCIEILTL+0PfTOIQcn2VPkpxMwf6Gbt9n4UEFDjZ4RuUKjd0\n-----END PRIVATE KEY-----",
            },
        ]
        issuer = PasetoIssuer(app)
        with pytest.raises(EncodeError) as err:
            issuer.issue({"foo": "bar"})
            pytest.fail("issue() must fail.")
        assert "Failed to encode a token." in str(err.value)



from typing import Optional

import pyseto
from flask import Flask, Request, Response
from flask.json.tag import TaggedJSONSerializer
from flask.sessions import SecureCookieSession, SessionInterface, SessionMixin
from pyseto import Key


class PasetoCookieSessionInterface(SessionInterface):
    def __init__(self, paseto_version: int = 4):
        if not isinstance(paseto_version, int) or paseto_version not in [1, 2, 3, 4]:
            raise ValueError(f"Invalid PASETO version: {paseto_version}")

        self._paseto_version = paseto_version
        self._serializer = TaggedJSONSerializer()

    def open_session(self, app: Flask, request: Request) -> Optional[SessionMixin]:
        """
        Opens a session.
        """
        if not app.secret_key:
            return None

        val = request.cookies.get(self.get_cookie_name(app))
        if not val:
            return SecureCookieSession()

        # max_age = int(app.permanent_session_lifetime.total_seconds())

        try:
            dec_key = Key.new(self._paseto_version, "local", app.secret_key)
            serialized_session = pyseto.decode(dec_key, val).payload.decode("utf-8")  # type: ignore
            return SecureCookieSession(self._serializer.loads(serialized_session))
        except Exception:
            return SecureCookieSession()

    def save_session(self, app: Flask, session: SessionMixin, response: Response) -> None:
        """
        Saves a session.
        """
        name = self.get_cookie_name(app)
        domain = self.get_cookie_domain(app)
        path = self.get_cookie_path(app)
        secure = self.get_cookie_secure(app)
        samesite = self.get_cookie_samesite(app)

        if not session:
            if session.modified:
                response.delete_cookie(name, domain=domain, path=path, secure=secure, samesite=samesite)
            return

        if session.accessed:
            response.vary.add("Cookie")

        if not self.should_set_cookie(app, session):
            return

        httponly = self.get_cookie_httponly(app)
        expires = self.get_expiration_time(app, session)
        enc_key = Key.new(self._paseto_version, "local", app.secret_key)
        serialized_session = self._serializer.dumps(dict(session))
        val = pyseto.encode(enc_key, serialized_session).decode("ascii")
        response.set_cookie(
            name,
            val,
            expires=expires,
            httponly=httponly,
            domain=domain,
            path=path,
            secure=secure,
            samesite=samesite,
        )
        return



# flake8: noqa: E501
import json

import flask
import pytest
from flask import jsonify, make_response

from flask_paseto_extended import (
    PasetoIssuer,
    PasetoVerifier,
    current_paseto,
    paseto_required,
)


@pytest.fixture(scope="function")
def app():
    # Mock user database.
    users = {"foo@bar.example": {"password": "mysecret"}}

    app = flask.Flask(__name__)

    # Configurations for PasetoIssuer.
    app.config["PASETO_ISS"] = "https://issuer.example"
    app.config["PASETO_USE_ISS"] = True
    app.config["PASETO_USE_IAT"] = True
    app.config["PASETO_EXP"] = 3600
    app.config["PASETO_PRIVATE_KEYS"] = [
        {
            "version": 4,
            "key": "-----BEGIN PRIVATE KEY-----\nMC4CAQAwBQYDK2VwBCIEILTL+0PfTOIQcn2VPkpxMwf6Gbt9n4UEFDjZ4RuUKjd0\n-----END PRIVATE KEY-----",
        },
    ]
    issuer = PasetoIssuer(app)

    # Configurations for PasetoVerifier.
    app.config["PASETO_SKEW"] = 60  # sec
    app.config["PASETO_PUBLIC_KEYS"] = [
        {
            "iss": "https://issuer.exmaple",
            "version": 4,
            "key": "-----BEGIN PUBLIC KEY-----\nMCowBQYDK2VwAyEAHrnbu7wEfAP9cGBOAHHwmH4Wsot1ciXBHwBBXQ4gsaI=\n-----END PUBLIC KEY-----",
        },
    ]
    verifier = PasetoVerifier(app)

    @verifier.token_loader
    def token_loader(request):
        return request.cookies.get("paseto", None)

    @verifier.verification_error_handler
    def verification_error_handler():
        resp = make_response("Unauthorized")
        resp.delete_cookie("paseto", httponly=True)
        return resp

    @app.route("/login", methods=["GET", "POST"])
    def login():
        email = flask.request.form["email"]
        if flask.request.form["password"] != users[email]["password"]:
            return "Bad login"

        token = issuer.issue(payload={"user": {"email": email}}).decode("utf-8")
        resp = flask.redirect(flask.url_for("protected"))
        resp.set_cookie("paseto", token, httponly=True)
        return resp

    @app.route("/logout")
    def logout():
        resp = make_response("Logged out")
        resp.delete_cookie("paseto", httponly=True)
        return resp

    @app.route("/protected")
    @paseto_required()
    def protected():
        assert current_paseto.is_verified is True
        assert current_paseto.version == "v4"
        assert current_paseto.purpose == "public"
        assert current_paseto.header == b"v4.public."
        assert current_paseto.footer == b""
        assert current_paseto.error is None
        return jsonify(current_paseto.payload["user"])

    return app


class TestPasetoIssuerAndVerifier:
    """
    Tests for sample code (examples/issuer_and_verifier.py).
    """

    def test_issuer_and_verifier(self, app):
        with app.test_client() as c:
            res = c.post(
                "/login",
                data={"email": "foo@bar.example", "password": "mysecret"},
                follow_redirects=True,
            )
            assert res.status_code == 200
            res = c.get("/protected")
            assert res.status_code == 200
            body = json.loads(res.data)
            assert body["email"] == "foo@bar.example"
            res = c.get("/logout", follow_redirects=True)
            assert res.status_code == 200

    def test_issuer_and_verifier_without_login(self, app):
        with app.test_client() as c:
            res = c.get("/protected")
            assert res.status_code == 200
            assert res.data == b"Unauthorized"



from datetime import timedelta
from secrets import token_bytes

import flask
import flask_login
import pytest
from werkzeug.http import parse_cookie

from flask_paseto_extended import PasetoLoginManager

# Mock database.
users = {"foo@bar.example": {"password": "mysecret"}}


# Simple user class.
class User(flask_login.UserMixin):
    pass


# Wrapper func of PasetoLoginManager().
def new_login_manager(app):
    login_manager = PasetoLoginManager(app)

    @login_manager.user_loader
    def user_loader(email):
        if email not in users:
            return
        user = User()
        user.id = email
        return user

    @login_manager.request_loader
    def request_loader(request):
        email = request.form.get("email")
        if email not in users:
            return
        user = User()
        user.id = email
        return user

    @login_manager.unauthorized_handler
    def unauthorized_handler():
        return "Unauthorized"

    return login_manager


@pytest.fixture(scope="function")
def app():
    app = flask.Flask(__name__)

    @app.route("/login", methods=["POST"])
    def login():
        email = flask.request.form["email"]
        if flask.request.form["password"] == users[email]["password"]:
            user = User()
            user.id = email
            flask_login.login_user(user, remember=True)
            return flask.redirect(flask.url_for("protected"))
        return "Bad login"

    @app.route("/logout")
    def logout():
        flask_login.logout_user()
        return "Logged out"

    @app.route("/protected")
    @flask_login.login_required
    def protected():
        return "Logged in as: " + flask_login.current_user.id

    return app


@pytest.fixture(scope="function")
def app_with_duration():
    app = flask.Flask(__name__)

    @app.route("/login", methods=["POST"])
    def login():
        email = flask.request.form["email"]
        if flask.request.form["password"] == users[email]["password"]:
            user = User()
            user.id = email
            flask_login.login_user(user, remember=True, duration=timedelta(seconds=1))
            return flask.redirect(flask.url_for("protected"))
        return "Bad login"

    @app.route("/logout")
    def logout():
        flask_login.logout_user()
        return "Logged out"

    @app.route("/protected")
    @flask_login.login_required
    def protected():
        return "Logged in as: " + flask_login.current_user.id

    return app


class TestPasetoLoginManager:
    """
    Tests for PasetoLoginManager.
    """

    def test_login_manager(self, app):
        app.secret_key = "super secret string"
        login_manager = new_login_manager(app)

        assert login_manager.paseto_version == 4

        with app.test_client() as c:
            res = c.post(
                "/login",
                data={"email": "foo@bar.example", "password": "mysecret"},
                follow_redirects=True,
            )
            assert res.status_code == 200
            res = c.get("/protected")
            assert res.status_code == 200
            assert res.data == b"Logged in as: foo@bar.example"
            res = c.get("/logout", follow_redirects=True)
            assert res.status_code == 200

    def test_login_manager_with_duration1(self, app):
        app.config["REMEMBER_COOKIE_DURATION"] = 1
        app.secret_key = "super secret string"
        new_login_manager(app)

        with app.test_client() as c:
            res = c.post(
                "/login",
                data={"email": "foo@bar.example", "password": "mysecret"},
                follow_redirects=True,
            )
            assert res.status_code == 200
            res = c.get("/protected")
            assert res.status_code == 200
            assert res.data == b"Logged in as: foo@bar.example"
            res = c.get("/logout", follow_redirects=True)
            assert res.status_code == 200

    def test_login_manager_with_duration2(self, app_with_duration):
        # app.config["REMEMBER_COOKIE_DURATION"] = 1
        app_with_duration.secret_key = "super secret string"
        new_login_manager(app_with_duration)

        with app_with_duration.test_client() as c:
            res = c.post(
                "/login",
                data={"email": "foo@bar.example", "password": "mysecret"},
                follow_redirects=True,
            )
            assert res.status_code == 200
            res = c.get("/protected")
            assert res.status_code == 200
            assert res.data == b"Logged in as: foo@bar.example"
            res = c.get("/logout", follow_redirects=True)
            assert res.status_code == 200

    def test_login_manager_with_invalid_duration(self, app):
        app.config["REMEMBER_COOKIE_DURATION"] = "1"
        app.secret_key = "super secret string"
        new_login_manager(app)

        with app.test_client() as c:
            res = c.post(
                "/login",
                data={"email": "foo@bar.example", "password": "mysecret"},
                follow_redirects=True,
            )
            assert res.status_code == 500

    def test_login_manager_with_refresh(self, app):
        app.config["REMEMBER_COOKIE_DURATION"] = 1
        app.secret_key = "super secret string"
        new_login_manager(app)

        with app.test_client() as c:
            res = c.post(
                "/login",
                data={"email": "foo@bar.example", "password": "mysecret"},
                # follow_redirects=True,
            )
            assert res.status_code == 302
            cookie = res.headers.getlist("Set-Cookie")[0]
            remember_token = parse_cookie(cookie)["remember_token"]

        class client:
            def __init__(self, app):
                self.app = app

            def __call__(self, environ, start_response):
                environ["HTTP_COOKIE"] = environ.get("HTTP_COOKIE", f"remember_token={remember_token}")
                return self.app(environ, start_response)

        app.wsgi_app = client(app.wsgi_app)
        with app.test_client() as c:
            res = c.get("/protected")
            assert res.status_code == 200
            assert res.data == b"Logged in as: foo@bar.example"

    @pytest.mark.parametrize(
        "version, key",
        [
            (1, "super secret string"),
            (2, token_bytes(32)),
            (3, "super secret string"),
            (4, "super secret string"),
        ],
    )
    def test_login_manager_with_paseto_version(self, app, version, key):
        app.secret_key = key
        app.config["REMEMBER_COOKIE_PASETO_VERSION"] = version

        login_manager = new_login_manager(app)

        assert login_manager.paseto_version == version
        assert login_manager._remember_cookie_key is None

        with app.test_client() as c:
            res = c.post(
                "/login",
                data={"email": "foo@bar.example", "password": "mysecret"},
                follow_redirects=True,
            )
            assert res.status_code == 200
            res = c.get("/protected")
            assert res.status_code == 200
            assert res.data == b"Logged in as: foo@bar.example"
            res = c.get("/logout", follow_redirects=True)
            assert res.status_code == 200

    def test_login_manager_with_remember_cookie_key(self, app):
        app.secret_key = "super secret string"
        app.config["REMEMBER_COOKIE_PASETO_KEY"] = "my super secret"
        login_manager = new_login_manager(app)

        assert login_manager.paseto_version == 4

        with app.test_client() as c:
            res = c.post(
                "/login",
                data={"email": "foo@bar.example", "password": "mysecret"},
                follow_redirects=True,
            )
            assert res.status_code == 200
            res = c.get("/protected")
            assert res.status_code == 200
            assert res.data == b"Logged in as: foo@bar.example"
            res = c.get("/logout", follow_redirects=True)
            assert res.status_code == 200

    @pytest.mark.parametrize(
        "version, msg",
        [
            (0, "REMEMBER_COOKIE_PASETO_VERSION must be 1, 2, 3 or 4"),
            (5, "REMEMBER_COOKIE_PASETO_VERSION must be 1, 2, 3 or 4"),
            (100, "REMEMBER_COOKIE_PASETO_VERSION must be 1, 2, 3 or 4"),
        ],
    )
    def test_login_manager_with_unsupported_version(self, app, version, msg):
        app.secret_key = "super secret string"
        app.config["REMEMBER_COOKIE_PASETO_VERSION"] = version

        with pytest.raises(ValueError) as err:
            PasetoLoginManager(app)
            pytest.fail("PasetoLoginManager() should fail.")
        assert msg in str(err.value)

    @pytest.mark.parametrize(
        "version, msg",
        [
            ("xxx", "REMEMBER_COOKIE_PASETO_VERSION must be int"),
            ({}, "REMEMBER_COOKIE_PASETO_VERSION must be int"),
            ([], "REMEMBER_COOKIE_PASETO_VERSION must be int"),
        ],
    )
    def test_login_manager_with_invalid_version(self, app, version, msg):
        app.secret_key = "super secret string"
        app.config["REMEMBER_COOKIE_PASETO_VERSION"] = version

        with pytest.raises(TypeError) as err:
            PasetoLoginManager(app)
            pytest.fail("PasetoLoginManager() should fail.")
        assert msg in str(err.value)

    @pytest.mark.parametrize(
        "key, msg",
        [
            (12345, "REMEMBER_COOKIE_PASETO_KEY must be str"),
            (b"my secret", "REMEMBER_COOKIE_PASETO_KEY must be str"),
            ({}, "REMEMBER_COOKIE_PASETO_KEY must be str"),
            ([], "REMEMBER_COOKIE_PASETO_KEY must be str"),
        ],
    )
    def test_login_manager_with_invalid_type_of_key(self, app, key, msg):
        app.secret_key = "super secret string"
        app.config["REMEMBER_COOKIE_PASETO_KEY"] = key

        with pytest.raises(TypeError) as err:
            PasetoLoginManager(app)
            pytest.fail("PasetoLoginManager() should fail.")
        assert msg in str(err.value)

    def test_login_manager_without_secret_key(self, app):
        # app.secret_key = "super secret string"
        new_login_manager(app)

        with app.test_client() as c:
            res = c.post(
                "/login",
                data={"email": "foo@bar.example", "password": "mysecret"},
                follow_redirects=True,
            )
            assert res.status_code == 500

    def test_login_manager_with_invalid_key(self, app):
        app.secret_key = "not 32bytes secret for v2"
        app.config["REMEMBER_COOKIE_PASETO_VERSION"] = 2
        new_login_manager(app)

        with app.test_client() as c:
            res = c.post(
                "/login",
                data={"email": "foo@bar.example", "password": "mysecret"},
                follow_redirects=True,
            )
            assert res.status_code == 500
